/*======================================================================

    NinjaSCSI-3R / NinjaSCSI-32Bi PCMCIA SCSI host adapter card driver
      By: YOKOTA Hiroshi <yokota{at}netlab.cs.tsukuba.ac.jp>

    Ver.3.0   Use kernel thread
    Ver.2.8   Support 32bit MMIO mode
              Support Synchronous Data Transfer Request (SDTR) mode
    Ver.2.0   Support 32bit PIO mode
    Ver.1.1.2 Fix for scatter list buffer exceeds
    Ver.1.1   Support scatter list
    Ver.0.1   Initial version

    This software may be used and distributed according to the terms of
    the GNU General Public License.

======================================================================*/

/***********************************************************************
    This driver is for these PCcards.

	I-O DATA PCSC-F	 (Workbit NinjaSCSI-3)
			"WBT", "NinjaSCSI-3", "R1.0"
	I-O DATA CBSC-II (Workbit NinjaSCSI-32Bi in 16bit mode)
			"IO DATA", "CBSC16	 ", "1"

***********************************************************************/

/* $Id: nsp_cs.c,v 1.59 2006/09/20 09:23:51 elca Exp $ */

#include <linux/blkdev.h>
#include <linux/completion.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/version.h>

#include <linux/moduleparam.h>

#include <asm/io.h>
#include <asm/irq.h>
#include <asm/semaphore.h>

#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_eh.h>
#include <scsi/scsi_tcq.h>
#include <scsi/scsi.h>

#include <scsi/scsi_ioctl.h>
#include <scsi/scsi_host.h>

#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>

#include "nsp_cs.h"

MODULE_DESCRIPTION("WorkBit NinjaSCSI-3 / NinjaSCSI-32Bi(16bit) PCMCIA SCSI host adapter module $Revision: 1.59 $");
MODULE_AUTHOR("YOKOTA Hiroshi <yokota{at}netlab.cs.tsukuba.ac.jp>");
MODULE_SUPPORTED_DEVICE("sd,sr,sg,st");
MODULE_LICENSE("GPL");

#include "nsp_io.h"

/*====================================================================*/
/* Parameters that can be set with 'insmod' */

static int      nsp_burst_mode = BURST_MEM32; /* default: memory mapped IO, 32 bit mode */
module_param    (nsp_burst_mode, int, S_IRUGO | S_IWUSR); /* -rw-r--r-- */
MODULE_PARM_DESC(nsp_burst_mode, "Burst transfer mode (0=io8, 1=io32, 2=mem32(default))");


static CLASS_DEVICE_ATTR(nsp_version,		  S_IRUGO, nsp_show_version,		 NULL);
static CLASS_DEVICE_ATTR(nsp_host_no,		  S_IRUGO, nsp_show_host_no,		 NULL);
static CLASS_DEVICE_ATTR(nsp_irq,		  S_IRUGO, nsp_show_irq,		 NULL);
static CLASS_DEVICE_ATTR(nsp_io_port,		  S_IRUGO, nsp_show_io_port,		 NULL);
static CLASS_DEVICE_ATTR(nsp_mmio_port,		  S_IRUGO, nsp_show_mmio_port,		 NULL);
static CLASS_DEVICE_ATTR(nsp_sg_tablesize,	  S_IRUGO, nsp_show_sg_tablesize,	 NULL);
static CLASS_DEVICE_ATTR(nsp_burst_transfer_mode, S_IRUGO, nsp_show_burst_transfer_mode, NULL);
static CLASS_DEVICE_ATTR(nsp_chip_id,		  S_IRUGO, nsp_show_chip_id,		 NULL);
static CLASS_DEVICE_ATTR(nsp_chip_revision,	  S_IRUGO, nsp_show_chip_revision,	 NULL);
static CLASS_DEVICE_ATTR(nsp_current_sc,	  S_IRUGO, nsp_show_current_sc,		 NULL);
static CLASS_DEVICE_ATTR(nsp_status,		  S_IRUGO, nsp_show_status,		 NULL);
static struct class_device_attribute *nsp_sysfs_attrs[] = {
	&class_device_attr_nsp_version,
	&class_device_attr_nsp_host_no,
	&class_device_attr_nsp_irq,
	&class_device_attr_nsp_io_port,
	&class_device_attr_nsp_mmio_port,
	&class_device_attr_nsp_sg_tablesize,
	&class_device_attr_nsp_burst_transfer_mode,
	&class_device_attr_nsp_chip_id,
	&class_device_attr_nsp_chip_revision,
	&class_device_attr_nsp_current_sc,
	&class_device_attr_nsp_status,
	NULL,
};

/* /usr/src/linux/include/scsi/scsi_host.h */
static struct scsi_host_template nsp_driver_template = {
	.module			 = THIS_MODULE,
	.proc_name		 = "nsp_cs",
	.proc_info		 = nsp_proc_info,
	.name			 = "WorkBit NinjaSCSI-3/32Bi(16bit)",
	.info			 = nsp_info,
	.queuecommand		 = nsp_queuecommand,
/*	.eh_strategy_handler	 = nsp_eh_strategy,*/
/*	.eh_abort_handler	 = nsp_eh_abort,*/
/*	.eh_device_reset_handler = nsp_eh_device_reset,*/
	.eh_bus_reset_handler	 = nsp_eh_bus_reset,
	.eh_host_reset_handler	 = nsp_eh_host_reset,
	.can_queue		 = 1,
	.this_id		 = NSP_INITIATOR_ID,
	.sg_tablesize		 = SG_ALL,
	.cmd_per_lun		 = 1,
	.use_clustering		 = DISABLE_CLUSTERING,
	.shost_attrs		 = nsp_sysfs_attrs,
};

static nsp_hw_data nsp_data_base; /* attach <-> detect glue */



/******************************************************************
 * debug, error print
 */
#ifdef NSP_DEBUG
# include "nsp_debug.c"
#endif	/* NSP_DEBUG */

#ifndef NSP_DEBUG
# define NSP_DEBUG_MASK		0x000000
# define nsp_msg(type, args...) nsp_cs_message(NULL, 0, (type), args)
# define nsp_dbg(mask, args...) do {} while(0)
#else
# define NSP_DEBUG_MASK		0xffffff
# define nsp_msg(type, args...) \
	nsp_cs_message (__FUNCTION__, __LINE__, (type), args)
# define nsp_dbg(mask, args...) \
	nsp_cs_dmessage(__FUNCTION__, __LINE__, (mask), args)
#endif

#define NSP_DEBUG_QUEUECOMMAND		BIT(0)
#define NSP_DEBUG_REGISTER		BIT(1)
#define NSP_DEBUG_AUTOSCSI		BIT(2)
#define NSP_DEBUG_INTR			BIT(3)
#define NSP_DEBUG_SGLIST		BIT(4)
#define NSP_DEBUG_BUSFREE		BIT(5)
#define NSP_DEBUG_CDB_CONTENTS		BIT(6)
#define NSP_DEBUG_RESELECTION		BIT(7)
#define NSP_DEBUG_MSGINOCCUR		BIT(8)
#define NSP_DEBUG_EEPROM		BIT(9)
#define NSP_DEBUG_MSGOUTOCCUR		BIT(10)
#define NSP_DEBUG_BUSRESET		BIT(11)
#define NSP_DEBUG_RESTART		BIT(12)
#define NSP_DEBUG_SYNC			BIT(13)
#define NSP_DEBUG_WAIT			BIT(14)
#define NSP_DEBUG_TARGETFLAG		BIT(15)
#define NSP_DEBUG_PROC			BIT(16)
#define NSP_DEBUG_INIT			BIT(17)
#define NSP_DEBUG_DATA_IO		BIT(18)
#define NSP_DEBUG_THREAD		BIT(19)
#define NSP_SPECIAL_PRINT_REGISTER	BIT(20)

#define NSP_DEBUG_BUF_LEN		150

__attribute__((format(printf, 4, 5)))
static void nsp_cs_message(const char *func, int line, const char *type, const char *fmt, ...)
{
	va_list args;
	char buf[NSP_DEBUG_BUF_LEN];

	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);

#ifndef NSP_DEBUG
	printk("%snsp_cs: %s\n", type, buf);
#else
	printk("%snsp_cs: %s (%d): %s\n", type, func, line, buf);
#endif
}

#ifdef NSP_DEBUG
__attribute__((format(printf, 4, 5)))
static void nsp_cs_dmessage(const char *func, int line, int mask, const char *fmt, ...)
{
	va_list args;
	char buf[NSP_DEBUG_BUF_LEN];

	va_start(args, fmt);
	vsnprintf(buf, sizeof(buf), fmt, args);
	va_end(args);

	if (mask & NSP_DEBUG_MASK) {
		printk("nsp_cs-debug: 0x%x %s (%d): %s\n", mask, func, line, buf);
	}
}
#endif

/***********************************************************/

/*====================================================
 * Clenaup parameters and call done() functions.
 * You must be set SCpnt->result before call this function.
 */
static void nsp_scsi_done(struct scsi_cmnd *SCpnt)
{
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;

	SCpnt->SCp.buffer = NULL;
	SCpnt->SCp.ptr    = NULL;
	SCpnt->SCp.this_residual = 0;
	SCpnt->SCp.buffers_residual = 0;
	SCpnt->SCp.Status	= SAM_STAT_CHECK_CONDITION;
	SCpnt->SCp.Message	= COMMAND_COMPLETE;
	SCpnt->SCp.have_data_in = IO_UNKNOWN;
	SCpnt->SCp.sent_command = 0;
	SCpnt->SCp.phase	= PH_UNDETERMINED;

	data->CurrentSC = NULL;

	SCpnt->scsi_done(SCpnt);
}

static int nsp_queuecommand(struct scsi_cmnd *SCpnt, void (*done)(struct scsi_cmnd *))
{
#ifdef NSP_DEBUG
	/*const unsigned int host_id = SCpnt->device->host->this_id;*/
	/*const unsigned int base    = SCpnt->device->host->io_port;*/
	const unsigned char target = scmd_id(SCpnt);
#endif
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;

	nsp_dbg(NSP_DEBUG_QUEUECOMMAND,
		"SCpnt=0x%p target=%d lun=%d buff=0x%p bufflen=%d use_sg=%d",
		SCpnt, target, SCpnt->device->lun,
		SCpnt->request_buffer, SCpnt->request_bufflen,
		SCpnt->use_sg);
	//nsp_dbg(NSP_DEBUG_QUEUECOMMAND, "before CurrentSC=0x%p", data->CurrentSC);

	SCpnt->scsi_done = done;

	if (data->CurrentSC     != NULL) {
		nsp_msg(KERN_WARNING, "CurrentSC!=NULL this can't be happen");
		SCpnt->result   = DID_BAD_TARGET << 16;
		nsp_scsi_done(SCpnt);
		return 0;
	}

	/* check target ID is not same as this initiator ID */
	if (scmd_id(SCpnt) == SCpnt->device->host->this_id) {
		nsp_dbg(NSP_DEBUG_QUEUECOMMAND, "terget==host???");
		SCpnt->result = DID_BAD_TARGET << 16;
		nsp_scsi_done(SCpnt);
		return 0;
	}

	show_command(SCpnt);

	data->CurrentSC		= SCpnt;

	/*--------------------------------
	 * fill suitable values to SCpnt */
	SCpnt->SCp.Status	= SAM_STAT_CHECK_CONDITION;
	SCpnt->SCp.Message	= COMMAND_COMPLETE;
	SCpnt->SCp.have_data_in = IO_UNKNOWN;
	SCpnt->SCp.sent_command = 0;
	SCpnt->SCp.phase	= PH_UNDETERMINED;
	SCpnt->resid		= SCpnt->request_bufflen;

	/* setup scratch area
	   SCp.ptr		: buffer pointer
	   SCp.this_residual	: buffer length
	   SCp.buffer		: next buffer
	   SCp.buffers_residual : left buffers in list
	   SCp.phase		: current state of the command */
	if (SCpnt->use_sg) {
		SCpnt->SCp.buffer	    = (struct scatterlist *) SCpnt->request_buffer;
		SCpnt->SCp.ptr		    = SG_ADDRESS(SCpnt->SCp.buffer);
		SCpnt->SCp.this_residual    = SCpnt->SCp.buffer->length;
		SCpnt->SCp.buffers_residual = SCpnt->use_sg - 1;
	} else {
		SCpnt->SCp.ptr		    = (char *) SCpnt->request_buffer;
		SCpnt->SCp.this_residual    = SCpnt->request_bufflen;
		SCpnt->SCp.buffer	    = NULL;
		SCpnt->SCp.buffers_residual = 0;
	}

	/* start selection phase */
	if (nsphw_start_selection(SCpnt) == FALSE) {
		nsp_dbg(NSP_DEBUG_QUEUECOMMAND, "selection fail");
		SCpnt->result   = DID_BUS_BUSY << 16;
		nsp_scsi_done(SCpnt);
		return 0;
	}

#ifdef NSP_DEBUG
	data->CmdId++;
#endif

	//nsp_dbg(NSP_DEBUG_QUEUECOMMAND, "out");
	return 0;
}

/*
 * setup PIO FIFO transfer mode and enable/disable to data out
 */
static void nsp_setup_fifo(const nsp_hw_data *data, int enabled)
{
	const unsigned int  base = data->BaseAddress;
	unsigned char transfer_mode_reg;

	//nsp_dbg(NSP_DEBUG_DATA_IO, "enabled=%d", enabled);

	if (enabled != FALSE) {
		transfer_mode_reg = TRANSFER_GO | BRAIND;
	} else {
		transfer_mode_reg = 0;
	}

	transfer_mode_reg |= data->TransferMode;

	nsp_index_write(base, TRANSFERMODE, transfer_mode_reg);
}

static void nsphw_init_sync(nsp_hw_data *data)
{
	const sync_data tmp_sync = {
		.SyncNegotiation = SYNC_NOT_YET,
		.SyncPeriod      = 0,
		.SyncOffset      = 0
	};
	int i;

	/* setup sync data */
	for ( i = 0; i < ARRAY_SIZE(data->Sync); i++ ) {
		data->Sync[i] = tmp_sync;
	}
}

/*
 * Initialize Ninja hardware
 */
static int nsphw_init(nsp_hw_data *data)
{
	const unsigned int base = data->BaseAddress;

	nsp_dbg(NSP_DEBUG_INIT, "in base=0x%x", base);

	data->ScsiClockDiv = CLOCK_20M /*CLOCK_40M | FAST_20*/;
	data->CurrentSC    = NULL;
	data->FifoCount    = 0;
	data->TransferMode = MODE_IO8;

	nsphw_init_sync(data);


	/* block all interrupts */
	nsp_write(base,	      IRQCONTROL,   IRQCONTROL_ALL_CLEAR_AND_MASK);

	nsp_write(base,	      IFSELECT,	    0);
	data->ChipRev = nsp_read(base, FIFOSTATUS);

	/* setup SCSI interface */
	nsp_write(base,	      IFSELECT,	    IF_REGSEL);

	nsp_index_write(base, SCSIIRQMODE,  0);

	nsp_index_write(base, TRANSFERMODE, MODE_IO8);
	nsp_index_write(base, CLOCKDIV,	    data->ScsiClockDiv);

	nsp_index_write(base, PARITYCTRL,   0);
	nsp_index_write(base, POINTERCLR,   POINTER_CLEAR     |
					    ACK_COUNTER_CLEAR |
					    REQ_COUNTER_CLEAR |
					    HOST_COUNTER_CLEAR);

	/* setup fifo asic */
	nsp_write(base,	      IFSELECT,	    IF_REGSEL);
	nsp_index_write(base, TERMPWRCTRL,  0);
	if ((nsp_index_read(base, OTHERCONTROL) & TPWR_SENSE) == 0) {
		nsp_msg(KERN_INFO, "terminator power on");
		nsp_index_write(base, TERMPWRCTRL, POWER_ON);
	}

	nsp_index_write(base, TIMERCOUNT,   0);
	nsp_index_write(base, TIMERCOUNT,   0); /* requires 2 times!! */

	nsp_index_write(base, SYNCREG,	    0);
	nsp_index_write(base, ACKWIDTH,	    0);

	/* enable interrupts and ack them */
	nsp_index_write(base, SCSIIRQMODE,  SCSI_PHASE_CHANGE_EI |
					    RESELECT_EI		 |
					    FIFO_IRQ_EI		 |
					    SCSI_RESET_IRQ_EI	 );
	nsp_write(base,	      IRQCONTROL,   IRQCONTROL_ALL_CLEAR);

	nsp_setup_fifo(data, FALSE);

	return TRUE;
}

/*
 * Start selection phase
 */
static int nsphw_start_selection(struct scsi_cmnd *SCpnt)
{
	const unsigned int  host_id = SCpnt->device->host->this_id;
	const unsigned int  base    = SCpnt->device->host->io_port;
	const unsigned char target  = scmd_id(SCpnt);
	nsp_hw_data  *data = (nsp_hw_data *)SCpnt->device->host->hostdata;
	unsigned char phase, arbit;
	unsigned long timeout_time;

	//nsp_dbg(NSP_DEBUG_RESELECTION, "in");

	phase = nsp_index_read(base, SCSIBUSMON);
	if(phase != BUSMON_BUS_FREE) {
		nsp_msg(KERN_INFO, "bus busy, target=%d", target);
		return FALSE;
	}

	/* start arbitration */
	//nsp_dbg(NSP_DEBUG_RESELECTION, "start arbit");
	SCpnt->SCp.phase = PH_ARBSTART;
	nsp_index_write(base, SETARBIT, ARBIT_GO);

	timeout_time = jiffies + msecs_to_jiffies(NSP_ARBIT_TIMEOUT);
	do {
		/* XXX: what a stupid chip! */
		arbit = nsp_index_read(base, ARBITSTATUS);
		//nsp_dbg(NSP_DEBUG_RESELECTION, "arbit=%d, wait_count=%d", arbit, wait_count);
		udelay(2); /* hold 1.2us */
	} while((arbit & (ARBIT_WIN | ARBIT_FAIL)) == 0 &&
		time_before(jiffies, timeout_time));

	if (!(arbit & ARBIT_WIN)) {
		//nsp_dbg(NSP_DEBUG_RESELECTION, "arbit fail");
		nsp_index_write(base, SETARBIT, ARBIT_FLAG_CLEAR);
		return FALSE;
	}

	/* assert select line */
	//nsp_dbg(NSP_DEBUG_RESELECTION, "assert SEL line");
	SCpnt->SCp.phase = PH_SELSTART;
	udelay(3); /* wait 2.4us */
	nsp_index_write(base, SCSIDATALATCH, BIT(host_id) | BIT(target));
	nsp_index_write(base, SCSIBUSCTRL,   SCSI_SEL | SCSI_BSY                    | SCSI_ATN);
	udelay(2); /* wait >1.2us */
	nsp_index_write(base, SCSIBUSCTRL,   SCSI_SEL | SCSI_BSY | SCSI_DATAOUT_ENB | SCSI_ATN);
	nsp_index_write(base, SETARBIT,	     ARBIT_FLAG_CLEAR);
	ndelay(90); /* wait >90ns */
	nsp_index_write(base, SCSIBUSCTRL,   SCSI_SEL		 | SCSI_DATAOUT_ENB | SCSI_ATN);

	/* check selection timeout */
	nsp_start_timer(SCpnt, 1000/51);
	data->SelectionTimeOut = 1;

	return TRUE;
}

/***********************************************************************
 * Period/AckWidth speed conversion table
 *
 * Note: This period/ackwidth speed table must be in descending order.
 ***********************************************************************/
struct nsp_sync_table {
	const unsigned int chip_period;
	const unsigned int ack_width;
	const unsigned int min_period;
	const unsigned int max_period;
};

const static struct nsp_sync_table nsp_sync_table_40M[] = {
     /* {PNo, AW,   SP,   EP}  Speed(MB/s) Period AckWidth */
	{0x1,  0, 0x0c, 0x0c},  /*  20.0 :  50ns,  25ns */
	{0x2,  0, 0x0d, 0x18},  /*  13.3 :  75ns,  25ns */
	{0x3,  1, 0x19, 0x19},  /*  10.0 : 100ns,  50ns */
	{0x4,  1, 0x1a, 0x1f},  /*   8.0 : 125ns,  50ns */
	{0x5,  2, 0x20, 0x25},  /*   7.5 : 150ns,  75ns */
	{0x6,  2, 0x26, 0x31},  /*   5.71: 175ns,  75ns */
	{0x7,  3, 0x32, 0x32},  /*   5.0 : 200ns, 100ns */
	{0x8,  3, 0x33, 0x38},  /*   4.44: 225ns, 100ns */
	{0x9,  3, 0x39, 0x3e},  /*   4.0 : 250ns, 100ns */
	{0xa,  3, 0x3f, 0x44},  /*   3.64: 275ns, 100ns */
	{0xb,  3, 0x45, 0x4b},  /*   3.33: 300ns, 100ns */
	{0xc,  3, 0x4c, 0x53},  /*   3.01: 325ns, 100ns */
	{0xd,  3, 0x54, 0x57},  /*   2.86: 350ns, 100ns */
	{0xe,  3, 0x58, 0x5d},  /*   2.67: 375ns, 100ns */
	{0xf,  3, 0x5e, 0x64},  /*   2.5 : 400ns, 100ns */
	{0,0,0,0},
};

const static struct nsp_sync_table nsp_sync_table_20M[] = {
	{0x1,  0, 0x19, 0x19},  /* 10.0 : 100ns,  50ns */
	{0x2,  0, 0x1a, 0x25},  /*  6.7 : 150ns,  50ns */
	{0x3,  1, 0x26, 0x32},  /*  5.0 : 200ns, 100ns */
	{0x4,  1, 0x33, 0x3e},  /*  4.0 : 250ns, 100ns */
	{0x5,  2, 0x3f, 0x4b},  /*  3.3 : 300ns, 150ns */
	{0x6,  2, 0x4c, 0x57},  /*  2.8 : 350ns, 150ns */
	{0x7,  3, 0x58, 0x64},  /*  2.5 : 400ns, 200ns */
	{0x8,  3, 0x65, 0x70},  /*  2.2 : 450ns, 200ns */
	{0x9,  3, 0x71, 0x7d},  /*  2.0 : 500ns, 200ns */
	{0xa,  3, 0x7e, 0x89},  /*  1.82: 550ns, 200ns */
	{0xb,  3, 0x8a, 0x95},  /*  1.67: 550ns, 200ns */
	{0xc,  3, 0x96, 0xa2},  /*  1.54: 550ns, 200ns */
	{0xd,  3, 0xa3, 0xae},  /*  1.43: 550ns, 200ns */
	{0xe,  3, 0xaf, 0xbb},  /*  1.33: 550ns, 200ns */
	{0xf,  3, 0xbc, 0xc8},  /*  1.25: 550ns, 200ns */
	{0,0,0,0},
};

/*
 * setup synchronous data transfer mode
 */
static int nsp_analyze_sdtr(const struct scsi_cmnd *SCpnt)
{
	const unsigned char    target = scmd_id(SCpnt);
//	const unsigned char    lun    = SCpnt->device->lun;
	nsp_hw_data	      *data   = (nsp_hw_data *)SCpnt->device->host->hostdata;
	sync_data	      *sync   = &(data->Sync[target]);
	const struct nsp_sync_table *sync_table;
	unsigned int	       period, offset;
	int		       i;


	nsp_dbg(NSP_DEBUG_SYNC, "in");

	period = sync->SyncPeriod;
	offset = sync->SyncOffset;

	nsp_dbg(NSP_DEBUG_SYNC, "period=0x%x, offset=0x%x", period, offset);

	switch (data->ScsiClockDiv) {
	case CLOCK_20M:
	case CLOCK_40M:
		sync_table = nsp_sync_table_20M;
		break;
	case (CLOCK_40M | FAST_20):
		sync_table = nsp_sync_table_40M;
		break;
	default:
		nsp_msg(KERN_WARNING,
			"Invalid clock div is selected, set 20M.");
		sync_table = nsp_sync_table_20M;
		break;
	}

	for ( i = 0; sync_table->max_period != 0; i++, sync_table++) {
		if ( period >= sync_table->min_period &&
		     period <= sync_table->max_period	 ) {
			break;
		}
	}

	if (period != 0 && sync_table->max_period == 0) {
		/*
		 * No proper period/offset found
		 */
		nsp_dbg(NSP_DEBUG_SYNC, "no proper period/offset");

		sync->SyncPeriod      = 0;
		sync->SyncOffset      = 0;
		sync->SyncRegister    = 0;
		sync->AckWidth	      = 0;

		return FALSE;
	}

	sync->SyncRegister    = (sync_table->chip_period << SYNCREG_PERIOD_SHIFT) |
				(offset & SYNCREG_OFFSET_MASK);
	sync->AckWidth	      = sync_table->ack_width;

	nsp_dbg(NSP_DEBUG_SYNC, "sync_reg=0x%x, ack_width=0x%x", sync->SyncRegister, sync->AckWidth);

	return TRUE;
}


/*
 * start ninja hardware timer
 */
static void nsp_start_timer(const struct scsi_cmnd *SCpnt, int time)
{
	const unsigned int base = SCpnt->device->host->io_port;
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;

	//nsp_dbg(NSP_DEBUG_INTR, "in SCpnt=0x%p, time=%d", SCpnt, time);
	data->TimerCount = time;
	nsp_index_write(base, TIMERCOUNT, time);
}

/*
 * wait for bus phase change
 */
static int nsp_negate_signal(const struct scsi_cmnd *SCpnt, unsigned char mask, char *str)
{
	const unsigned int  base = SCpnt->device->host->io_port;
	unsigned char reg;
	unsigned long timeout_time;

	//nsp_dbg(NSP_DEBUG_INTR, "in");

	timeout_time = jiffies + msecs_to_jiffies(NSP_SIGNAL_TIMEOUT);

	do {
		reg = nsp_index_read(base, SCSIBUSMON);
		if (reg == 0xff) {
			nsp_msg(KERN_DEBUG, "bus monitor err");
			break;
		}
	} while ( ( (reg & mask) != 0 ) &&
		  time_before(jiffies, timeout_time));

	if (time_after_eq(jiffies, timeout_time)) {
		nsp_msg(KERN_DEBUG, " %s signal off timeut", str);
	}

	return 0;
}

/*
 * expect Ninja Irq
 */
static int nsp_expect_signal(const struct scsi_cmnd	   *SCpnt,
			     unsigned char  current_phase,
			     unsigned char  mask)
{
	const unsigned int  base = SCpnt->device->host->io_port;
	unsigned long	    timeout_time;

	//nsp_dbg(NSP_DEBUG_INTR, "current_phase=0x%x, mask=0x%x", current_phase, mask);

	timeout_time = jiffies + msecs_to_jiffies(NSP_SIGNAL_TIMEOUT);
	do {
		unsigned char phase, i_src;

		phase = nsp_index_read(base, SCSIBUSMON);
		if (phase == 0xff) {
			//nsp_dbg(NSP_DEBUG_INTR, "ret -1");
			return -1;
		}
		i_src = nsp_read(base, IRQSTATUS);
		if (i_src & IRQSTATUS_SCSI) {
			//nsp_dbg(NSP_DEBUG_INTR, "ret 0 found scsi signal");
			return 0;
		}
		if ((phase & mask) != 0 && (phase & BUSMON_PHASE_MASK) == current_phase) {
			//nsp_dbg(NSP_DEBUG_INTR, "ret 1 phase=0x%x", phase);
			return 1;
		}
	} while(time_before(jiffies, timeout_time));

	nsp_msg(KERN_DEBUG, "expect signal timeout");
	return -1;
}

/*
 * get extra SCSI data from fifo
 */
static int nsp_dataphase_bypass(struct scsi_cmnd *SCpnt)
{
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;
	unsigned int count;

	//nsp_dbg(NSP_DEBUG_DATA_IO, "in");

	if (SCpnt->SCp.have_data_in != IO_IN) {
		return 0;
	}

	count = nsp_fifo_count(SCpnt);
	if (data->FifoCount == count) {
		//nsp_dbg(NSP_DEBUG_DATA_IO, "not use bypass quirk");
		return 0;
	}

	/*
	 * XXX: NSP_QUIRK
	 * data phase skip only occures in case of SCSI_LOW_READ
	 */
	nsp_dbg(NSP_DEBUG_DATA_IO, "use bypass quirk");
	SCpnt->SCp.phase = PH_DATA;
	nsp_pio_read(SCpnt);
	nsp_setup_fifo(data, FALSE);

	return 0;
}

/*
 * accept reselection
 */
static int nsp_reselected(const struct scsi_cmnd *SCpnt)
{
	const unsigned int  base    = SCpnt->device->host->io_port;
	const unsigned int  host_id = SCpnt->device->host->this_id;
	//nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;
	unsigned char bus_reg;
	unsigned char id_reg, tmp;
	int target;

	nsp_dbg(NSP_DEBUG_RESELECTION, "in");

	id_reg = nsp_index_read(base, RESELECTID);
	tmp    = id_reg & (~BIT(host_id));
	target = 0;
	while(tmp != 0) {
		if (tmp & BIT(0)) {
			break;
		}
		tmp >>= 1;
		target++;
	}

	if (scmd_id(SCpnt) != target) {
		nsp_msg(KERN_ERR, "XXX: reselect ID must be %d in this implementation.", target);
	}

	nsp_negate_signal(SCpnt, BUSMON_SEL, "reselect<SEL>");

	nsp_nexus(SCpnt);
	bus_reg = nsp_index_read(base, SCSIBUSCTRL) & ~(SCSI_BSY | SCSI_ATN);
	nsp_index_write(base, SCSIBUSCTRL, bus_reg);
	nsp_index_write(base, SCSIBUSCTRL, bus_reg | AUTODIRECTION | ACKENB);

	return TRUE;
}

/*
 * count how many data transferd
 */
static int nsp_fifo_count(const struct scsi_cmnd *SCpnt)
{
	const unsigned int base = SCpnt->device->host->io_port;
	unsigned int count;
	unsigned int l, m, h, dummy;

	nsp_index_write(base, POINTERCLR, POINTER_CLEAR | ACK_COUNTER);

	l     = nsp_index_read(base, TRANSFERCOUNT);
	m     = nsp_index_read(base, TRANSFERCOUNT);
	h     = nsp_index_read(base, TRANSFERCOUNT);
	dummy = nsp_index_read(base, TRANSFERCOUNT); /* required this! */

	count = (h << 16) | (m << 8) | (l << 0);

	//nsp_dbg(NSP_DEBUG_DATA_IO, "count=0x%x", count);

	return count;
}

/* fifo size */
#define RFIFO_CRIT 64
#define WFIFO_CRIT 64

/*
 * read data in DATA IN phase
 */
static void nsp_pio_read(struct scsi_cmnd *SCpnt)
{
	const unsigned int  base      = SCpnt->device->host->io_port;
	const unsigned long mmio_base = SCpnt->device->host->base;
	nsp_hw_data  *data      = (nsp_hw_data *)SCpnt->device->host->hostdata;
	int	      ocount;
	unsigned long timeout_time;

	ocount = data->FifoCount;

	nsp_dbg(NSP_DEBUG_DATA_IO, "in SCpnt=0x%p resid=0x%x ocount=%d ptr=0x%p this_residual=0x%x buffers=0x%p nbuf=%d",
		SCpnt, SCpnt->resid, ocount, SCpnt->SCp.ptr, SCpnt->SCp.this_residual, SCpnt->SCp.buffer, SCpnt->SCp.buffers_residual);
	timeout_time = jiffies + msecs_to_jiffies(NSP_RW_TIMEOUT);

	while (time_before(jiffies, timeout_time) &&
	       (SCpnt->SCp.this_residual > 0 || SCpnt->SCp.buffers_residual > 0 ) ) {
		unsigned char stat, fifo_stat;
		int res;

		stat = nsp_index_read(base, SCSIBUSMON);
		stat &= BUSMON_PHASE_MASK;


		res = nsp_fifo_count(SCpnt) - ocount;
		//nsp_dbg(NSP_DEBUG_DATA_IO, "ptr=0x%p this=0x%x ocount=0x%x res=0x%x", SCpnt->SCp.ptr, SCpnt->SCp.this_residual, ocount, res);
		if (res == 0) { /* if some data avilable ? */
			if (stat == BUSPHASE_DATA_IN) { /* phase changed? */
				//nsp_dbg(NSP_DEBUG_DATA_IO, " wait for data this=%d", SCpnt->SCp.this_residual);
				continue;
			} else {
				nsp_dbg(NSP_DEBUG_DATA_IO, "phase changed stat=0x%x", stat);
				break;
			}
		}

		fifo_stat = nsp_read(base, FIFOSTATUS);
		if ((fifo_stat & FIFOSTATUS_FULL_EMPTY) == 0 &&
		    stat				== BUSPHASE_DATA_IN) {
			continue;
		}

		res = min(res, SCpnt->SCp.this_residual);

		switch (data->TransferMode) {
		case MODE_IO32:
			res &= ~(BIT(1)|BIT(0)); /* align 4 */
			nsp_fifo32_read(base, SCpnt->SCp.ptr, res >> 2);
			break;
		case MODE_IO8:
			nsp_fifo8_read (base, SCpnt->SCp.ptr, res     );
			break;

		case MODE_MEM32:
			res &= ~(BIT(1)|BIT(0)); /* align 4 */
			nsp_mmio_fifo32_read(mmio_base, SCpnt->SCp.ptr, res >> 2);
			break;

		default:
			nsp_dbg(NSP_DEBUG_DATA_IO, "unknown read mode");
			return;
		}

		SCpnt->resid		 -= res;
		SCpnt->SCp.ptr		 += res;
		SCpnt->SCp.this_residual -= res;
		ocount			 += res;
		//nsp_dbg(NSP_DEBUG_DATA_IO, "ptr=0x%p this_residual=0x%x ocount=0x%x", SCpnt->SCp.ptr, SCpnt->SCp.this_residual, ocount);

		/* go to next scatter list if available */
		if (SCpnt->SCp.this_residual	== 0 &&
		    SCpnt->SCp.buffers_residual != 0 ) {
			//nsp_dbg(NSP_DEBUG_DATA_IO, "scatterlist next timeout_time=%ld", timeout_time);
			SCpnt->SCp.buffers_residual--;
			SCpnt->SCp.buffer++;
			SCpnt->SCp.ptr		 = SG_ADDRESS(SCpnt->SCp.buffer);
			SCpnt->SCp.this_residual = SCpnt->SCp.buffer->length;
			timeout_time = jiffies + msecs_to_jiffies(NSP_RW_TIMEOUT);

			//nsp_dbg(NSP_DEBUG_DATA_IO, "page: 0x%p, off: 0x%x", SCpnt->SCp.buffer->page, SCpnt->SCp.buffer->offset);
		}
	}

	data->FifoCount = ocount;

	if (time_after_eq(jiffies, timeout_time)) {
		nsp_msg(KERN_DEBUG, "pio read timeout resid=0x%x this_residual=0x%x buffers_residual=%d",
			SCpnt->resid, SCpnt->SCp.this_residual, SCpnt->SCp.buffers_residual);
	}
	nsp_dbg(NSP_DEBUG_DATA_IO, "read ocount=0x%x jiffies=%ld timeout_time=%ld", ocount, jiffies, timeout_time);
	nsp_dbg(NSP_DEBUG_DATA_IO, "r cmd=%d resid=0x%x\n", data->CmdId, SCpnt->resid);

#ifdef NSP_DEBUG
	if (SCpnt->cmnd[0] == 0x12) {
		nsp_byte_dump(SCpnt->SCp.ptr, 0, SCpnt->request_bufflen);
	}
#endif
}

/*
 * write data in DATA OUT phase
 */
static void nsp_pio_write(struct scsi_cmnd *SCpnt)
{
	const unsigned int  base      = SCpnt->device->host->io_port;
	const unsigned long mmio_base = SCpnt->device->host->base;
	nsp_hw_data  *data      = (nsp_hw_data *)SCpnt->device->host->hostdata;
	int	      ocount;
	unsigned long timeout_time;

	ocount	 = data->FifoCount;

	nsp_dbg(NSP_DEBUG_DATA_IO, "in fifocount=%d ptr=0x%p this_residual=%d buffers=0x%p nbuf=%d resid=0x%x",
		data->FifoCount, SCpnt->SCp.ptr, SCpnt->SCp.this_residual, SCpnt->SCp.buffer, SCpnt->SCp.buffers_residual, SCpnt->resid);

	timeout_time = jiffies + msecs_to_jiffies(NSP_RW_TIMEOUT);

	while (time_before(jiffies, timeout_time) &&
	       (SCpnt->SCp.this_residual > 0 || SCpnt->SCp.buffers_residual > 0)) {
		unsigned char stat;
		int res;

		stat = nsp_index_read(base, SCSIBUSMON);
		stat &= BUSMON_PHASE_MASK;

		if (stat != BUSPHASE_DATA_OUT) {
			res = ocount - nsp_fifo_count(SCpnt);

			nsp_dbg(NSP_DEBUG_DATA_IO, "phase changed stat=0x%x, res=%d\n", stat, res);
			/* Put back pointer */
			SCpnt->resid		 += res;
			SCpnt->SCp.ptr		 -= res;
			SCpnt->SCp.this_residual += res;
			ocount			 -= res;

			break;
		}

		res = ocount - nsp_fifo_count(SCpnt);
		if (res > 0) { /* write all data? */
			nsp_dbg(NSP_DEBUG_DATA_IO, "wait for all data out. ocount=0x%x res=%d", ocount, res);
			continue;
		}

		res = min(SCpnt->SCp.this_residual, WFIFO_CRIT);

		if(SCpnt->SCp.ptr == NULL) {
			nsp_msg(KERN_ERR, "ptr==NULL (w) buffers_residual=%d buffer=0x%p ptr=0x%p this_residual=%d length=%d",
				SCpnt->SCp.buffers_residual,
				SCpnt->SCp.buffer,
				SCpnt->SCp.ptr,
				SCpnt->SCp.this_residual,
				SCpnt->SCp.buffer->length);
			break;
		}

		//nsp_dbg(NSP_DEBUG_DATA_IO, "ptr=0x%p this=0x%x res=0x%x", SCpnt->SCp.ptr, SCpnt->SCp.this_residual, res);
		switch (data->TransferMode) {
		case MODE_IO32:
			res &= ~(BIT(1)|BIT(0)); /* align 4 */
			nsp_fifo32_write(base, SCpnt->SCp.ptr, res >> 2);
			break;
		case MODE_IO8:
			nsp_fifo8_write (base, SCpnt->SCp.ptr, res     );
			break;

		case MODE_MEM32:
			res &= ~(BIT(1)|BIT(0)); /* align 4 */
			nsp_mmio_fifo32_write(mmio_base, SCpnt->SCp.ptr, res >> 2);
			break;

		default:
			nsp_dbg(NSP_DEBUG_DATA_IO, "unknown write mode");
			break;
		}

		SCpnt->resid		 -= res;
		SCpnt->SCp.ptr		 += res;
		SCpnt->SCp.this_residual -= res;
		ocount			 += res;

		/* go to next scatter list if available */
		if (SCpnt->SCp.this_residual	== 0 &&
		    SCpnt->SCp.buffers_residual != 0 ) {
			nsp_dbg(NSP_DEBUG_DATA_IO, "scatterlist next");
			SCpnt->SCp.buffers_residual--;
			SCpnt->SCp.buffer++;
			SCpnt->SCp.ptr		 = SG_ADDRESS(SCpnt->SCp.buffer);
			SCpnt->SCp.this_residual = SCpnt->SCp.buffer->length;
			timeout_time = jiffies + msecs_to_jiffies(NSP_RW_TIMEOUT);
		}
	}

	data->FifoCount = ocount;

	if (time_after_eq(jiffies, timeout_time)) {
		nsp_msg(KERN_DEBUG, "pio write timeout resid=0x%x", SCpnt->resid);
	}
	nsp_dbg(NSP_DEBUG_DATA_IO, "write ocount=0x%x jiffies=%ld, timeout_time=%ld", ocount, jiffies, timeout_time);
	nsp_dbg(NSP_DEBUG_DATA_IO, "w cmd=%d resid=0x%x\n", data->CmdId, SCpnt->resid);
}
#undef RFIFO_CRIT
#undef WFIFO_CRIT

/*
 * setup synchronous/asynchronous data transfer mode
 */
static int nsp_nexus(const struct scsi_cmnd *SCpnt)
{
	const unsigned int   base   = SCpnt->device->host->io_port;
	const unsigned char  target = scmd_id(SCpnt);
//	const unsigned char  lun    = SCpnt->device->lun;
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;
	const sync_data     *sync   = &(data->Sync[target]);

	//nsp_dbg(NSP_DEBUG_DATA_IO, "in SCpnt=0x%p", SCpnt);

	/* setup synch transfer registers */
	nsp_index_write(base, SYNCREG,	sync->SyncRegister);
	nsp_index_write(base, ACKWIDTH, sync->AckWidth);

	if (SCpnt->use_sg    == 0	 ||
	    SCpnt->resid % 4 != 0	 ||
	    SCpnt->resid     <= PAGE_SIZE ) {
		data->TransferMode = MODE_IO8;
	} else if (nsp_burst_mode == BURST_MEM32) {
		data->TransferMode = MODE_MEM32;
	} else if (nsp_burst_mode == BURST_IO32) {
		data->TransferMode = MODE_IO32;
	} else if (nsp_burst_mode == BURST_IO8) {
		data->TransferMode = MODE_IO8;
	} else {
		nsp_msg(KERN_ERR,
			"Unknown burst mode. Fall back to IO8. use_sg=%d resid=%d",
			SCpnt->use_sg, SCpnt->resid);
		data->TransferMode = MODE_IO8;
	}

	/* setup pdma fifo */
	nsp_setup_fifo(data, TRUE);

	/* clear ack counter */
	data->FifoCount = 0;
	nsp_index_write(base, POINTERCLR, POINTER_CLEAR	    |
					  ACK_COUNTER_CLEAR |
					  REQ_COUNTER_CLEAR |
					  HOST_COUNTER_CLEAR);

	return 0;
}

#include "nsp_message.c"
/*
 * interrupt handler
 */
static irqreturn_t nspintr(int irq, void *dev_id, struct pt_regs *regs)
{
	unsigned int   base;
	unsigned char  irq_status, irq_phase, phase;
	struct scsi_cmnd     *tmpSC;
	struct Scsi_Host *host;
	unsigned char  target, lun;
	unsigned int  *sync_neg;
	int	       i, tmp;
	unsigned long  flags;
	nsp_hw_data   *data;
	int	       handled = 0;

	//nsp_dbg(NSP_DEBUG_INTR, "dev_id=0x%p", dev_id);

	if (dev_id != NULL) {
		host = (struct Scsi_Host *)dev_id;
		data = (nsp_hw_data *)(host->hostdata);
	} else {
		nsp_dbg(NSP_DEBUG_INTR, "host data wrong dev_id=0x%p", dev_id);
		return IRQ_NONE;
	}

	spin_lock_irqsave(host->host_lock, flags);

	//nsp_dbg(NSP_DEBUG_INTR, "&nsp_data_base=0x%p, dev_id=0x%p", &nsp_data_base, dev_id);

	base = data->BaseAddress;
	//nsp_dbg(NSP_DEBUG_INTR, "base=0x%x", base);

	/*
	 * interrupt check
	 */
	nsp_write(base, IRQCONTROL, IRQCONTROL_ALL_MASK);
	irq_status = nsp_read(base, IRQSTATUS);
	//nsp_dbg(NSP_DEBUG_INTR, "irq_status=0x%x", irq_status);
	if ((irq_status == 0xff) || ((irq_status & IRQSTATUS_MASK) == 0)) {
		//nsp_dbg(NSP_DEBUG_INTR, "no irq/shared irq status=0x%x", irq_status);
		goto out;
	}
	handled = 1;

	/* XXX: IMPORTANT
	 * Do not read an irq_phase register if no scsi phase interrupt.
	 * Unless, you should lose a scsi phase interrupt.
	 */
	phase = nsp_index_read(base, SCSIBUSMON);
	if((irq_status & IRQSTATUS_SCSI) != 0) {
		irq_phase = nsp_index_read(base, IRQPHASESENCE);
	} else {
		irq_phase = 0;
	}

	//nsp_dbg(NSP_DEBUG_INTR, "irq_phase=0x%x", irq_phase);

	/*
	 * timer interrupt handler (scsi vs timer interrupts)
	 */
	//nsp_dbg(NSP_DEBUG_INTR, "timercount=%d", data->TimerCount);
	if (data->TimerCount != 0) {
		//nsp_dbg(NSP_DEBUG_INTR, "stop timer");
		nsp_index_write(base, TIMERCOUNT, 0);
		nsp_index_write(base, TIMERCOUNT, 0);
		data->TimerCount = 0;
	}

	if ((irq_status & IRQSTATUS_MASK) == IRQSTATUS_TIMER &&
	    data->SelectionTimeOut == 0) {
		//nsp_dbg(NSP_DEBUG_INTR, "timer start");
		nsp_write(base, IRQCONTROL, IRQCONTROL_TIMER_CLEAR);
		goto out;
	}

	nsp_write(base, IRQCONTROL, IRQCONTROL_TIMER_CLEAR | IRQCONTROL_FIFO_CLEAR);

	if ((irq_status & IRQSTATUS_SCSI) &&
	    (irq_phase  & SCSI_RESET_IRQ)) {
		nsp_msg(KERN_ERR, "bus reset (power off?)");

		nsphw_init(data);
		nsp_bus_reset(data);

		if(data->CurrentSC != NULL) {
			tmpSC = data->CurrentSC;
			tmpSC->result  = (DID_RESET		      << 16) |
					 ((tmpSC->SCp.Message & 0xff) <<  8) |
					 ((tmpSC->SCp.Status  & 0xff) <<  0);
			nsp_scsi_done(tmpSC);
		}
		goto out;
	}

	if (data->CurrentSC == NULL) {
		nsp_msg(KERN_ERR, "CurrentSC==NULL irq_status=0x%x phase=0x%x irq_phase=0x%x this can't be happen. reset everything", irq_status, phase, irq_phase);
		nsphw_init(data);
		nsp_bus_reset(data);
		goto out;
	}

	tmpSC    = data->CurrentSC;
	target   = scmd_id(tmpSC);
	lun      = tmpSC->device->lun;
	sync_neg = &(data->Sync[target].SyncNegotiation);

	/*
	 * parse hardware SCSI irq reasons register
	 */
	if (irq_status & IRQSTATUS_SCSI) {
		if (irq_phase & RESELECT_IRQ) {
			nsp_dbg(NSP_DEBUG_INTR, "reselect");
			nsp_write(base, IRQCONTROL, IRQCONTROL_RESELECT_CLEAR);
			if (nsp_reselected(tmpSC) != FALSE) {
				goto out;
			}
		}

		if ((irq_phase & (PHASE_CHANGE_IRQ | LATCHED_BUS_FREE)) == 0) {
			goto out;
		}
	}

	//show_phase(tmpSC);

	switch(tmpSC->SCp.phase) {
	case PH_SELSTART:
		// *sync_neg = SYNC_NOT_YET;
		if ((phase & BUSMON_BSY) == 0) {
			//nsp_dbg(NSP_DEBUG_INTR, "selection count=%d", data->SelectionTimeOut);
			if (data->SelectionTimeOut >= NSP_SELTIMEOUT_COUNT) {
				nsp_dbg(NSP_DEBUG_INTR, "selection time out");
				data->SelectionTimeOut = 0;
				nsp_index_write(base, SCSIBUSCTRL, 0);

				tmpSC->result   = DID_TIME_OUT << 16;
				nsp_scsi_done(tmpSC);

				goto out;
			}
			data->SelectionTimeOut++;
			nsp_start_timer(tmpSC, 1000/51);
			goto out;
		}

		/* attention assert */
		//nsp_dbg(NSP_DEBUG_INTR, "attention assert");
		data->SelectionTimeOut = 0;
		tmpSC->SCp.phase       = PH_SELECTED;
		nsp_index_write(base, SCSIBUSCTRL, SCSI_ATN);
		udelay(1);
		nsp_index_write(base, SCSIBUSCTRL, SCSI_ATN | AUTODIRECTION | ACKENB);
		goto out;

		break;

	case PH_RESELECT:
		//nsp_dbg(NSP_DEBUG_INTR, "phase reselect");
		// *sync_neg = SYNC_NOT_YET;
		if ((phase & BUSMON_PHASE_MASK) != BUSPHASE_MESSAGE_IN) {

			tmpSC->result	= DID_ABORT << 16;
			nsp_scsi_done(tmpSC);
			goto out;
		}
		/* fall thru */
	default:
		if ((irq_status & (IRQSTATUS_SCSI | IRQSTATUS_FIFO)) == 0) {
			goto out;
		}
		break;
	}

	/*
	 * SCSI sequencer
	 */
	//nsp_dbg(NSP_DEBUG_INTR, "start scsi seq");

	/* normal disconnect */
	if (((tmpSC->SCp.phase == PH_MSG_IN) || (tmpSC->SCp.phase == PH_MSG_OUT)) &&
	    (irq_phase & LATCHED_BUS_FREE) ) {
		nsp_dbg(NSP_DEBUG_INTR, "normal disconnect irq_status=0x%x, phase=0x%x, irq_phase=0x%x", irq_status, phase, irq_phase);

		// *sync_neg       = SYNC_NOT_YET;

		if ((tmpSC->SCp.Message == COMMAND_COMPLETE)) {     /* all command complete and return status */
			tmpSC->result = (DID_OK			     << 16) |
					((tmpSC->SCp.Message & 0xff) <<  8) |
					((tmpSC->SCp.Status  & 0xff) <<  0);
			nsp_dbg(NSP_DEBUG_INTR, "command complete result=0x%x", tmpSC->result);
			nsp_scsi_done(tmpSC);

			goto out;
		}

		goto out;
	}


	/* check unexpected bus free state */
	if (phase == 0) {
		nsp_msg(KERN_DEBUG, "unexpected bus free. irq_status=0x%x, phase=0x%x, irq_phase=0x%x", irq_status, phase, irq_phase);

		*sync_neg       = SYNC_NG;
		tmpSC->result   = (DID_ERROR << 16) | (tmpSC->SCp.Status << 0);
		nsp_scsi_done(tmpSC);
		goto out;
	}

	switch (phase & BUSMON_PHASE_MASK) {
	case BUSPHASE_COMMAND:
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE_COMMAND");
		if ((phase & BUSMON_REQ) == 0) {
			nsp_dbg(NSP_DEBUG_INTR, "REQ == 0");
			goto out;
		}

		tmpSC->SCp.phase = PH_COMMAND;

		nsp_nexus(tmpSC);

		/* write scsi command */
		nsp_dbg(NSP_DEBUG_INTR, "cmd_len=%d", tmpSC->cmd_len);
		nsp_index_write(base, COMMANDCTRL, CLEAR_COMMAND_POINTER);
		for (i = 0; i < tmpSC->cmd_len; i++) {
			nsp_index_write(base, COMMANDDATA, tmpSC->cmnd[i]);
		}
		nsp_index_write(base, COMMANDCTRL, CLEAR_COMMAND_POINTER | AUTO_COMMAND_GO);
		break;

	case BUSPHASE_DATA_OUT:
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE_DATA_OUT");

		tmpSC->SCp.phase	= PH_DATA;
		tmpSC->SCp.have_data_in = IO_OUT;

		nsp_pio_write(tmpSC);

		break;

	case BUSPHASE_DATA_IN:
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE_DATA_IN");

		tmpSC->SCp.phase	= PH_DATA;
		tmpSC->SCp.have_data_in = IO_IN;

		nsp_pio_read(tmpSC);

		break;

	case BUSPHASE_STATUS:
		nsp_dataphase_bypass(tmpSC);
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE_STATUS");

		tmpSC->SCp.phase = PH_STATUS;

		tmpSC->SCp.Status = nsp_index_read(base, SCSIDATAWITHACK);
		nsp_dbg(NSP_DEBUG_INTR, "message=0x%x status=0x%x", tmpSC->SCp.Message, tmpSC->SCp.Status);

		break;

	case BUSPHASE_MESSAGE_OUT:
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE_MESSAGE_OUT");
		if ((phase & BUSMON_REQ) == 0) {
			goto timer_out;
		}

		tmpSC->SCp.phase = PH_MSG_OUT;

		// *sync_neg = SYNC_NOT_YET;

		data->MsgOutLen = 0;
		nsp_build_identify(tmpSC);

		if (*sync_neg == SYNC_NOT_YET) {
			int period, offset;

			data->Sync[target].SyncPeriod = 0;
			data->Sync[target].SyncOffset = 0;

			/**/
			/* get max sync from sync table */
			switch (data->ScsiClockDiv) {
			case CLOCK_20M:
			case CLOCK_40M:
				period = nsp_sync_table_20M[0].max_period;
				offset = 15;
				break;
			case (CLOCK_40M|FAST_20):
				period = nsp_sync_table_40M[0].max_period;
				offset = 15;
				break;
			default:
				period = 0;
				offset = 0;
				break;
			}

			nsp_build_sdtr(tmpSC, period, offset);

			/**/
		}

		nsp_analyze_sdtr(tmpSC);
		show_message(data);
		nsp_message_out(tmpSC);
		break;

	case BUSPHASE_MESSAGE_IN:
		nsp_dataphase_bypass(tmpSC);
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE_MESSAGE_IN");
		if ((phase & BUSMON_REQ) == 0) {
			goto timer_out;
		}

		tmpSC->SCp.phase = PH_MSG_IN;
		data->MsgInLen = 0;
		nsp_message_in(tmpSC);

		/**/
		if (*sync_neg == SYNC_NOT_YET) {
			//nsp_dbg(NSP_DEBUG_INTR, "sync target=%d,lun=%d",target,lun);

			if (data->MsgInLen	 >= 5		     &&
			    data->MsgInBuffer[0] == EXTENDED_MESSAGE &&
			    data->MsgInBuffer[1] == 3		     &&
			    data->MsgInBuffer[2] == EXTENDED_SDTR    ) {
				data->Sync[target].SyncPeriod = data->MsgInBuffer[3];
				data->Sync[target].SyncOffset = data->MsgInBuffer[4];
				//nsp_dbg(NSP_DEBUG_INTR, "sync ok, %d %d", data->MsgInBuffer[3], data->MsgInBuffer[4]);
				*sync_neg = SYNC_OK;
			} else {
				data->Sync[target].SyncPeriod = 0;
				data->Sync[target].SyncOffset = 0;
				*sync_neg = SYNC_NG;
			}
			nsp_analyze_sdtr(tmpSC);
		}
		/**/

		/* search last messeage byte */
		tmp = -1;
		for (i = 0; i < data->MsgInLen; i++) {
			tmp = data->MsgInBuffer[i];
			if (data->MsgInBuffer[i] == EXTENDED_MESSAGE) {
				i += (1 + data->MsgInBuffer[i+1]);
			}
		}
		tmpSC->SCp.Message = tmp;

		nsp_dbg(NSP_DEBUG_INTR, "message=0x%x len=%d", tmpSC->SCp.Message, data->MsgInLen);
		show_message(data);

		break;

	case BUSPHASE_SELECT:
	default:
		nsp_dbg(NSP_DEBUG_INTR, "BUSPHASE other");

		break;
	}

	goto out;


timer_out:
	nsp_start_timer(tmpSC, 1000/102);
 out:
	nsp_write(base, IRQCONTROL, 0); /* clear IRQ mask */
	spin_unlock_irqrestore(host->host_lock, flags);
	//nsp_dbg(NSP_DEBUG_INTR, "out");
	return IRQ_RETVAL(handled);
}


/*----------------------------------------------------------------*/
/* look for ninja3 card and init if found			  */
/*----------------------------------------------------------------*/
static struct Scsi_Host *nsp_detect(struct scsi_host_template *sht)
{
	struct Scsi_Host *host;	/* registered host structure */
	nsp_hw_data      *data;

	nsp_dbg(NSP_DEBUG_INIT, "this_id=%d", sht->this_id);
	host = scsi_host_alloc(&nsp_driver_template, sizeof(nsp_hw_data));

	if (host == NULL) {
		nsp_dbg(NSP_DEBUG_INIT, "host failed");
		goto err;
	}

	/* Copy global variable to driver specific area */
	data  = (nsp_hw_data *)host->hostdata;
	*data = nsp_data_base;

	data->ScsiInfo->host = host;
#ifdef NSP_DEBUG
	data->CmdId = 0;
#endif

	nsp_dbg(NSP_DEBUG_INIT, "irq base=0x%p,%d data=0x%p,%d info=0x%p host=0x%p", &nsp_data_base, (&nsp_data_base)->IrqNumber, data, data->IrqNumber, data->ScsiInfo, data->ScsiInfo->host);

	host->unique_id	= data->BaseAddress;
	host->io_port   = data->BaseAddress;
	host->n_io_port	= data->NumAddress;
	host->irq       = data->IrqNumber;
	host->base      = data->MmioAddress;

	spin_lock_init(&(data->Lock));

	snprintf(data->nspinfo,
		 sizeof(data->nspinfo),
		 "NinjaSCSI-3/32Bi Driver $Revision: 1.59 $ IO:0x%04lx-0x%04lx MMIO(virt addr):0x%04lx IRQ:%02d",
		 host->io_port, host->io_port + host->n_io_port - 1,
		 host->base,
		 host->irq);
	sht->name = data->nspinfo;

	nsp_dbg(NSP_DEBUG_INIT, "end");

	return host; /* detect done. */


 err:
	return NULL;
}

/*----------------------------------------------------------------*/
/* return info string						  */
/*----------------------------------------------------------------*/
static const char *nsp_info(struct Scsi_Host *shpnt)
{
	const nsp_hw_data *data = (nsp_hw_data *)shpnt->hostdata;

	return data->nspinfo;
}

#undef SPRINTF
#define SPRINTF(args...) \
	do { \
		if(length > (pos - buffer)) { \
			pos += snprintf(pos, length - (pos - buffer) + 1, ## args); \
			nsp_dbg(NSP_DEBUG_PROC, "buffer=0x%p pos=0x%p length=%d %d\n", buffer, pos, length,  length - (pos - buffer));\
		} \
	} while(0)

/* Shows Ninja host card information for user. */
static int
nsp_proc_info(struct Scsi_Host *host,
	      char  *buffer,
	      char **start,
	      off_t  offset,
	      int    length,
	      int    inout)
{
	char		 *pos = buffer;
	int		  id;
	int		  thislength;
	unsigned long	  flags;
	nsp_hw_data	 *data;

	if (inout) {
		return -EINVAL;
	}

	data = (nsp_hw_data *)host->hostdata;


	SPRINTF("NinjaSCSI status\n\n");
	SPRINTF("Driver version:        $Revision: 1.59 $\n");
	SPRINTF("SCSI host No.:         %d\n",		host->host_no);
	SPRINTF("IRQ:                   %d\n",		host->irq);
	SPRINTF("IO:                    0x%lx-0x%lx\n", host->io_port, host->io_port + host->n_io_port - 1);
	SPRINTF("MMIO(virtual address): 0x%lx-0x%lx\n", host->base, host->base + data->MmioLength - 1);
	SPRINTF("sg_tablesize:          %d\n",		host->sg_tablesize);

	SPRINTF("burst transfer mode:   ");
	switch (nsp_burst_mode) {
	case BURST_IO8:
		SPRINTF("io8");
		break;
	case BURST_IO32:
		SPRINTF("io32");
		break;
	case BURST_MEM32:
		SPRINTF("mem32");
		break;
	default:
		SPRINTF("???");
		break;
	}
	SPRINTF("\n");
	SPRINTF("Chip ID:               %d\n", data->ChipRev >> 4);
	SPRINTF("Chip revision:         %d\n", data->ChipRev &  0x0f);


	spin_lock_irqsave(&(data->Lock), flags);
	SPRINTF("CurrentSC:             0x%p\n\n", data->CurrentSC);
	spin_unlock_irqrestore(&(data->Lock), flags);

	SPRINTF("SDTR status\n");
	for(id = 0; id < ARRAY_SIZE(data->Sync); id++) {

		SPRINTF("id %d: ", id);

		if (id == host->this_id) {
			SPRINTF("----- NinjaSCSI-3 host adapter\n");
			continue;
		}

		switch(data->Sync[id].SyncNegotiation) {
		case SYNC_OK:
			SPRINTF(" sync");
			break;
		case SYNC_NG:
			SPRINTF("async");
			break;
		case SYNC_NOT_YET:
			SPRINTF(" none");
			break;
		default:
			SPRINTF("?????");
			break;
		}

		if (data->Sync[id].SyncPeriod != 0) {
			int speed;
			speed = 1000000 / (data->Sync[id].SyncPeriod * 4);

			SPRINTF(" transfer %d.%dMB/s, offset %d",
				speed / 1000,
				speed % 1000,
				data->Sync[id].SyncOffset
				);
		}
		SPRINTF("\n");
	}

	thislength = pos - (buffer + offset);

	if(thislength < 0) {
		*start = 0;
		return 0;
	}


	thislength = min(thislength, length);
	*start = buffer + offset;

	return thislength;
}
#undef SPRINTF

/*---------------------------------------------------------------*/
/* error handler                                                 */
/*---------------------------------------------------------------*/

/*static int nsp_eh_strategy(struct Scsi_Host *Shost)
{
	return FAILED;
}*/

/*
static int nsp_eh_abort(struct scsi_cmnd *SCpnt)
{
	nsp_dbg(NSP_DEBUG_BUSRESET, "SCpnt=0x%p", SCpnt);

	return nsp_eh_bus_reset(SCpnt);
}*/

/*
static int nsp_eh_device_reset(struct scsi_cmnd *SCpnt)
{
	nsp_dbg(NSP_DEBUG_BUSRESET, "%s: SCpnt=0x%p", SCpnt);

	return FAILED;
}*/

/* Do bus reset. This function uses in low-level initialize functions. */
static int nsp_bus_reset(nsp_hw_data *data)
{
	const unsigned int base = data->BaseAddress;
	int i;

	nsp_msg(KERN_WARNING, "Bus reset");
	nsp_write(base, IRQCONTROL, IRQCONTROL_ALL_CLEAR_AND_MASK);

	nsp_index_write(base, SCSIBUSCTRL, SCSI_RST);
	mdelay(100); /* 100ms */
	nsp_index_write(base, SCSIBUSCTRL, 0);
	for(i = 0; i < 5; i++) {
		nsp_index_read(base, IRQPHASESENCE); /* dummy read */
	}

	nsphw_init_sync(data);

	nsp_write(base, IRQCONTROL, IRQCONTROL_ALL_CLEAR);

	if (data->CurrentSC != NULL) {
		nsp_msg(KERN_WARNING, "clean up current scsi command.");
		data->CurrentSC->result = DID_ERROR << 16;
		nsp_scsi_done(data->CurrentSC);
	}

	return SUCCESS;
}

/* Do bus reset. This function uses in high-level SCSI driver. */
static int nsp_eh_bus_reset(struct scsi_cmnd *SCpnt)
{
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;

	nsp_dbg(NSP_DEBUG_BUSRESET, "SCpnt=0x%p", SCpnt);

	return nsp_bus_reset(data);
}

/* Initialise Ninja host adapter. */
static int nsp_eh_host_reset(struct scsi_cmnd *SCpnt)
{
	nsp_hw_data *data = (nsp_hw_data *)SCpnt->device->host->hostdata;

	nsp_dbg(NSP_DEBUG_BUSRESET, "in");
	nsp_msg(KERN_DEBUG, "host reset");

	nsphw_init(data);

	return SUCCESS;
}


/**********************************************************************
  PCMCIA functions
**********************************************************************/

/*======================================================================
    nsp_cs_probe() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

    The dev_link structure is initialized, but we don't actually
    configure the card at this point -- we wait until we receive a
    card insertion event.
======================================================================*/
static int nsp_cs_probe(struct pcmcia_device *pdev)
{
	scsi_info_t  *info;
	nsp_hw_data  *data = &nsp_data_base;
	int ret;

	nsp_dbg(NSP_DEBUG_INIT, "in");

	/* Create new SCSI device */
	info = kmalloc(sizeof(*info), GFP_KERNEL);
	if (info == NULL) { return -ENOMEM; }
	memset(info, 0, sizeof(*info));
	info->p_dev = pdev;
	pdev->priv = info;
	data->ScsiInfo = info;

	nsp_dbg(NSP_DEBUG_INIT, "info=0x%p", info);

	/* The io structure describes IO port mapping */
	pdev->io.NumPorts1	 = 0x10;
	pdev->io.Attributes1	 = IO_DATA_PATH_WIDTH_AUTO;
	pdev->io.IOAddrLines	 = 10;	/* not used */

	/* Interrupt setup */
	pdev->irq.Attributes	 = IRQ_TYPE_EXCLUSIVE;
	pdev->irq.IRQInfo1	 = IRQ_LEVEL_ID;
	pdev->irq.Handler	 = NULL;
	pdev->irq.Instance       = NULL;

	/* General socket configuration */
	pdev->conf.Attributes	 = CONF_ENABLE_IRQ;
	pdev->conf.IntType	 = INT_MEMORY_AND_IO;
	pdev->conf.Present	 = PRESENT_OPTION;

	ret = nsp_cs_config(pdev);

	nsp_dbg(NSP_DEBUG_INIT, "pdev=0x%p", pdev);
	return ret;
} /* nsp_cs_probe */


/*======================================================================
    This deletes a driver "instance".  The device is de-registered
    with Card Services.	 If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.
======================================================================*/
static void nsp_cs_remove(struct pcmcia_device *pdev)
{
	scsi_info_t *info = pdev->priv;

	nsp_dbg(NSP_DEBUG_INIT, "in, pdev=0x%p", pdev);

	info->stop = 1;
	nsp_cs_release(pdev);

	kfree(pdev->priv);
	pdev->priv = NULL;
} /* nsp_cs_remove */


/*======================================================================
    nsp_cs_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet device available to the system.
======================================================================*/
#define CS_CHECK(fn, ret) \
do { last_fn = (fn); if ((last_ret = (ret)) != CS_SUCCESS) goto cs_failed; } while (0)
/*====================================================================*/
static int nsp_cs_config(struct pcmcia_device *pdev)
{
	scsi_info_t	 *info	 = pdev->priv;
	tuple_t		  tuple;
	cisparse_t	  parse;
	int		  last_ret, last_fn;
	unsigned char	  tuple_data[64];
	config_info_t	  conf;
	win_req_t	  req;
	memreq_t	  map;
	cistpl_cftable_entry_t dflt = { 0 };
	struct Scsi_Host *host;
	nsp_hw_data	 *data_b = &nsp_data_base;
	int		  error;

	nsp_dbg(NSP_DEBUG_INIT, "in info=0x%p", info);

	tuple.DesiredTuple    = CISTPL_CONFIG;
	tuple.Attributes      = 0;
	tuple.TupleData	      = tuple_data;
	tuple.TupleDataMax    = sizeof(tuple_data);
	tuple.TupleOffset     = 0;
	CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(pdev, &tuple	   ));
	CS_CHECK(GetTupleData,	pcmcia_get_tuple_data( pdev, &tuple	   ));
	CS_CHECK(ParseTuple,	pcmcia_parse_tuple(    pdev, &tuple, &parse));
	pdev->conf.ConfigBase = parse.config.base;
	pdev->conf.Present    = parse.config.rmask[0];

	/* Look up the current Vcc */
	CS_CHECK(GetConfigurationInfo, pcmcia_get_configuration_info(pdev, &conf));

	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	CS_CHECK(GetFirstTuple, pcmcia_get_first_tuple(pdev, &tuple));
	while (1) {
		cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);

		if (pcmcia_get_tuple_data(pdev, &tuple	) != 0 ||
		    pcmcia_parse_tuple(	  pdev, &tuple, &parse) != 0) {
			goto next_entry;
		}

		if (cfg->flags & CISTPL_CFTABLE_DEFAULT) { dflt = *cfg; }
		if (cfg->index == 0) { goto next_entry; }
		pdev->conf.ConfigIndex = cfg->index;

		/* Does this card need audio output? */
		if (cfg->flags & CISTPL_CFTABLE_AUDIO) {
			pdev->conf.Attributes |= CONF_ENABLE_SPKR;
			pdev->conf.Status = CCSR_AUDIO_ENA;
		}

		/* Use power settings for Vcc and Vpp if present */
		/*  Note that the CIS values need to be rescaled */
		if (cfg->vcc.present & (1 << CISTPL_POWER_VNOM)) {
			if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM] / 10000) {
				goto next_entry;
			}
		} else if (dflt.vcc.present & (1 << CISTPL_POWER_VNOM)) {
			if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM] / 10000) {
				goto next_entry;
			}
		}

		if (cfg->vpp1.present & (1 << CISTPL_POWER_VNOM)) {
			pdev->conf.Vpp =
				cfg->vpp1.param[CISTPL_POWER_VNOM] / 10000;
		} else if (dflt.vpp1.present & (1 << CISTPL_POWER_VNOM)) {
			pdev->conf.Vpp =
				dflt.vpp1.param[CISTPL_POWER_VNOM] / 10000;
		}

		/* Do we need to allocate an interrupt? */
		if (cfg->irq.IRQInfo1 || dflt.irq.IRQInfo1) {
			pdev->conf.Attributes |= CONF_ENABLE_IRQ;
		}

		/* IO window settings */
		pdev->io.NumPorts1 = pdev->io.NumPorts2 = 0;
		if ((cfg->io.nwin > 0) || (dflt.io.nwin > 0)) {
			cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt.io;
			pdev->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
			if (!(io->flags & CISTPL_IO_8BIT))
				pdev->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
			if (!(io->flags & CISTPL_IO_16BIT))
				pdev->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
			pdev->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
			pdev->io.BasePort1 = io->win[0].base;
			pdev->io.NumPorts1 = io->win[0].len;
			if (io->nwin > 1) {
				pdev->io.Attributes2 = pdev->io.Attributes1;
				pdev->io.BasePort2   = io->win[1].base;
				pdev->io.NumPorts2   = io->win[1].len;
			}
			/* This reserves IO space but doesn't actually enable it */
			if (pcmcia_request_io(pdev, &pdev->io) != 0) {
				goto next_entry;
			}
		}

		if ((cfg->mem.nwin > 0) || (dflt.mem.nwin > 0)) {
			cistpl_mem_t *mem =
				(cfg->mem.nwin) ? &cfg->mem : &dflt.mem;
			req.Attributes = WIN_DATA_WIDTH_16 | WIN_MEMORY_TYPE_CM;
			req.Attributes |= WIN_ENABLE;
			req.Base = mem->win[0].host_addr;
			req.Size = mem->win[0].len;
			if (req.Size < 0x1000) {
				req.Size = 0x1000;
			}
			req.AccessSpeed = 0;
			if (pcmcia_request_window(&pdev, &req, &pdev->win) != 0) {
				goto next_entry;
			}
			map.Page = 0; map.CardOffset = mem->win[0].card_addr;
			if (pcmcia_map_mem_page(pdev->win, &map) != 0) {
				goto next_entry;
			}

			data_b->MmioAddress = (unsigned long)ioremap_nocache(req.Base, req.Size);
			data_b->MmioLength  = req.Size;
		}
		/* If we got this far, we're cool! */
		break;

	next_entry:
		nsp_dbg(NSP_DEBUG_INIT, "next");
		CS_CHECK(GetNextTuple, pcmcia_get_next_tuple(pdev, &tuple));
	}

	if (pdev->conf.Attributes & CONF_ENABLE_IRQ) {
		CS_CHECK(RequestIRQ, pcmcia_request_irq(pdev, &pdev->irq));
	}
	CS_CHECK(RequestConfiguration, pcmcia_request_configuration(pdev, &pdev->conf));


	/* Set port and IRQ */
	data_b->BaseAddress = pdev->io.BasePort1;
	data_b->NumAddress  = pdev->io.NumPorts1;
	data_b->IrqNumber   = pdev->irq.AssignedIRQ;

	nsp_dbg(NSP_DEBUG_INIT, "I/O[0x%x+0x%x] IRQ %d",
		data_b->BaseAddress, data_b->NumAddress, data_b->IrqNumber);

	if(nsphw_init(data_b) == FALSE) {
		goto cs_failed;
	}

	/* nsp_data_base is not userble after nsp_detect() is called */
	host = nsp_detect(&nsp_driver_template);

	if (host == NULL) {
		nsp_dbg(NSP_DEBUG_INIT, "detect failed");
		goto cs_failed;
	}

	info->host = host;
	snprintf(info->node.dev_name, sizeof(info->node.dev_name), "scsi%d", host->host_no);
	pdev->dev_node  = &(info->node);

	error = request_irq(pdev->irq.AssignedIRQ, nspintr, /*IRQF_SHARED |*/ IRQF_SAMPLE_RANDOM, "nsp_cs", host);
	if (error != 0) {
		nsp_msg(KERN_ERR, "request_irq %d failed", pdev->irq.AssignedIRQ);
		goto cs_failed;
	}

	error = scsi_add_host (host, NULL);
	if (error != 0) {
		nsp_msg(KERN_ERR, "Add host fail");
		goto free_irq;
	}
	scsi_scan_host(host);

	/* Finally, report what we've done */
	printk(KERN_INFO "nsp_cs: index 0x%02x:",
	       pdev->conf.ConfigIndex);
	if (pdev->conf.Vpp) {
		printk(" Vpp %d.%d", pdev->conf.Vpp / 10, pdev->conf.Vpp % 10);
	}
	if (pdev->conf.Attributes & CONF_ENABLE_IRQ) {
		printk(" irq %d", pdev->irq.AssignedIRQ);
	}
	if (pdev->io.NumPorts1) {
		printk(" io 0x%04x-0x%04x", pdev->io.BasePort1,
		       pdev->io.BasePort1 + pdev->io.NumPorts1 - 1);
	}
	if (pdev->io.NumPorts2)
		printk(" & 0x%04x-0x%04x", pdev->io.BasePort2,
		       pdev->io.BasePort2 + pdev->io.NumPorts2 - 1);
	if (pdev->win)
		printk(" mem 0x%06lx-0x%06lx", req.Base,
		       req.Base+req.Size - 1);
	printk("\n");

	return 0;

 free_irq:
	free_irq(pdev->irq.AssignedIRQ, host);

 cs_failed:
	nsp_dbg(NSP_DEBUG_INIT, "config fail");
	cs_error(pdev, last_fn, last_ret);
	nsp_cs_release(pdev);

	return -ENODEV;
} /* nsp_cs_config */
#undef CS_CHECK


/*======================================================================
    After a card is removed, nsp_cs_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
======================================================================*/
static void nsp_cs_release(struct pcmcia_device *pdev)
{
	scsi_info_t *info = pdev->priv;
	nsp_hw_data *data = NULL;

	if (info->host == NULL) {
		nsp_msg(KERN_DEBUG, "unexpected card release call.");
	} else {
		data = (nsp_hw_data *)info->host->hostdata;
	}

	nsp_dbg(NSP_DEBUG_INIT, "pdev=0x%p", pdev);

	/* Unlink the device chain */
	if (info->host != NULL) {
		scsi_remove_host(info->host);
		free_irq(pdev->irq.AssignedIRQ, info->host);
	}

	pdev->dev_node = NULL;

	if (pdev->win) {
		if (data != NULL) {
			iounmap((void *)(data->MmioAddress));
		}
	}
	pcmcia_disable_device(pdev);

	if (info->host != NULL) {
		scsi_host_put(info->host);
	}
} /* nsp_cs_release */

static int nsp_cs_suspend(struct pcmcia_device *pdev)
{
	scsi_info_t *info = pdev->priv;
	nsp_hw_data *data;

	nsp_dbg(NSP_DEBUG_INIT, "event: suspend");

	if (info->host != NULL) {
		nsp_msg(KERN_INFO, "clear SDTR status");

		data = (nsp_hw_data *)info->host->hostdata;

		nsphw_init_sync(data);
	}

	info->stop = 1;

	return 0;
}

static int nsp_cs_resume(struct pcmcia_device *pdev)
{
	scsi_info_t *info = pdev->priv;
	nsp_hw_data *data;

	nsp_dbg(NSP_DEBUG_INIT, "event: resume");

	info->stop = 0;

	if (info->host != NULL) {
		nsp_msg(KERN_INFO, "reset host and bus");

		data = (nsp_hw_data *)info->host->hostdata;

		nsphw_init   (data);
		nsp_bus_reset(data);
	}

	return 0;
}

/************************************************************************
 * sysfs entries
 */
static ssize_t nsp_show_version(struct class_device *dev, char *buf)
{
	return snprintf(buf, PAGE_SIZE, "$Revision: 1.59 $\n");
}

static ssize_t nsp_show_host_no(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);

	return snprintf(buf, PAGE_SIZE, "%d\n", host->host_no);
}

static ssize_t nsp_show_irq(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);

	return snprintf(buf, PAGE_SIZE, "%d\n", host->irq);
}

static ssize_t nsp_show_io_port(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);

	return snprintf(buf, PAGE_SIZE, "0x%lx-0x%lx\n", host->io_port, host->io_port + host->n_io_port - 1);
}

static ssize_t nsp_show_mmio_port(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);
	const nsp_hw_data      *data = (nsp_hw_data *)host->hostdata;

	return snprintf(buf, PAGE_SIZE, "0x%lx-0x%lx\n", host->base, host->base + data->MmioLength - 1);
}

static ssize_t nsp_show_sg_tablesize(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);

	return snprintf(buf, PAGE_SIZE, "%d\n",	host->sg_tablesize);
}

static ssize_t nsp_show_burst_transfer_mode(struct class_device *dev, char *buf)
{
	char *mode_str;

	switch (nsp_burst_mode) {
	case BURST_IO8:
		mode_str = "io8";
		break;
	case BURST_IO32:
		mode_str = "io32";
		break;
	case BURST_MEM32:
		mode_str = "mem32";
		break;
	default:
		mode_str = "???";
		break;
	}

	return snprintf(buf, PAGE_SIZE, "%s\n", mode_str);
}

static ssize_t nsp_show_chip_id(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);
	const nsp_hw_data      *data = (nsp_hw_data *)host->hostdata;

	return snprintf(buf, PAGE_SIZE, "%d\n", data->ChipRev >> 4);
}

static ssize_t nsp_show_chip_revision(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);
	const nsp_hw_data      *data = (nsp_hw_data *)host->hostdata;

	return snprintf(buf, PAGE_SIZE, "%d\n", data->ChipRev &  0x0f);
}

static ssize_t nsp_show_current_sc(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);
	const nsp_hw_data      *data = (nsp_hw_data *)host->hostdata;
	unsigned long		flags;
	struct scsi_cmnd       *tmpsc;

	spin_lock_irqsave(&(data->Lock), flags);
	tmpsc = data->CurrentSC;
	spin_unlock_irqrestore(&(data->Lock), flags);

	return snprintf(buf, PAGE_SIZE, "0x%p", tmpsc);
}

#undef SPRINTF
#define SPRINTF(args...) \
	do { \
		if(PAGE_SIZE > (pos - buf)) { \
			pos += snprintf(pos, PAGE_SIZE - (pos - buf), ## args); \
			nsp_dbg(NSP_DEBUG_PROC, "buf=0x%p pos=0x%p length=%ld %ld\n", buf, pos, PAGE_SIZE,  PAGE_SIZE - (pos - buf));\
		} \
	} while(0)
static ssize_t nsp_show_status(struct class_device *dev, char *buf)
{
	const struct Scsi_Host *host = class_to_shost(dev);
	nsp_hw_data	       *data = (nsp_hw_data *)host->hostdata;
	char		 *pos = buf;
	unsigned long	  flags;
	int		  id;

	SPRINTF("NinjaSCSI status\n\n");
	SPRINTF("Driver version:        $Revision: 1.59 $\n");
	SPRINTF("SCSI host No.:         %d\n",		host->host_no);
	SPRINTF("IRQ:                   %d\n",		host->irq);
	SPRINTF("IO:                    0x%lx-0x%lx\n", host->io_port, host->io_port + host->n_io_port - 1);
	SPRINTF("MMIO(virtual address): 0x%lx-0x%lx\n", host->base, host->base + data->MmioLength - 1);
	SPRINTF("sg_tablesize:          %d\n",		host->sg_tablesize);

	SPRINTF("burst transfer mode:   ");
	switch (nsp_burst_mode) {
	case BURST_IO8:
		SPRINTF("io8");
		break;
	case BURST_IO32:
		SPRINTF("io32");
		break;
	case BURST_MEM32:
		SPRINTF("mem32");
		break;
	default:
		SPRINTF("???");
		break;
	}
	SPRINTF("\n");
	SPRINTF("Chip ID:               %d\n", data->ChipRev >> 4);
	SPRINTF("Chip revision:         %d\n", data->ChipRev &  0x0f);


	spin_lock_irqsave(&(data->Lock), flags);
	SPRINTF("CurrentSC:             0x%p\n\n", data->CurrentSC);
	spin_unlock_irqrestore(&(data->Lock), flags);

	SPRINTF("SDTR status\n");
	for(id = 0; id < ARRAY_SIZE(data->Sync); id++) {

		SPRINTF("id %d: ", id);

		if (id == host->this_id) {
			SPRINTF("----- NinjaSCSI-3 host adapter\n");
			continue;
		}

		switch(data->Sync[id].SyncNegotiation) {
		case SYNC_OK:
			SPRINTF(" sync");
			break;
		case SYNC_NG:
			SPRINTF("async");
			break;
		case SYNC_NOT_YET:
			SPRINTF(" none");
			break;
		default:
			SPRINTF("?????");
			break;
		}

		if (data->Sync[id].SyncPeriod != 0) {
			int speed;
			speed = 1000000 / (data->Sync[id].SyncPeriod * 4);

			SPRINTF(" transfer %d.%dMB/s, offset %d",
				speed / 1000,
				speed % 1000,
				data->Sync[id].SyncOffset
				);
		}
		SPRINTF("\n");
	}

	return pos - buf;
}
#undef SPRINTF

/*======================================================================*
 *	module entry point
 *====================================================================*/
static struct pcmcia_device_id nsp_cs_ids[] = {
	/* NinjaSCSI-3R */
	PCMCIA_DEVICE_PROD_ID123("WBT",     "NinjaSCSI-3",   "R1.0", 0xc7ba805f, 0xfdc7c97d, 0x6973710e),

	/* NinjaSCSI-32Bi */
	PCMCIA_DEVICE_PROD_ID123("IO DATA", "CBSC16       ", "1",    0x547e66dc, 0x0d63a3fd, 0x51de003a),
	PCMCIA_DEVICE_PROD_ID123("KME    ", "SCSI-CARD-001", "1",    0x534c02bc, 0x52008408, 0x51de003a),
	PCMCIA_DEVICE_PROD_ID123("KME    ", "SCSI-CARD-002", "1",    0x534c02bc, 0xcb09d5b2, 0x51de003a),
	PCMCIA_DEVICE_PROD_ID123("KME    ", "SCSI-CARD-003", "1",    0x534c02bc, 0xbc0ee524, 0x51de003a),
	PCMCIA_DEVICE_PROD_ID123("KME    ", "SCSI-CARD-004", "1",    0x534c02bc, 0x226a7087, 0x51de003a),
	PCMCIA_DEVICE_PROD_ID123("WORKBIT", "UltraNinja-16", "1",    0x28191418, 0xb70f4b09, 0x51de003a),
	PCMCIA_DEVICE_NULL,
};
MODULE_DEVICE_TABLE(pcmcia, nsp_cs_ids);

static struct pcmcia_driver nsp_driver = {
	.owner		= THIS_MODULE,
	.drv		= {
		.name	= "nsp_cs",
	},
	.probe		= nsp_cs_probe,
	.remove		= nsp_cs_remove,
	.id_table	= nsp_cs_ids,
	.suspend	= nsp_cs_suspend,
	.resume		= nsp_cs_resume,
};

static int __init nsp_cs_init(void)
{
	nsp_msg(KERN_INFO, "loading...");

	return pcmcia_register_driver(&nsp_driver);
}

static void __exit nsp_cs_exit(void)
{
	nsp_msg(KERN_INFO, "unloading...");

	pcmcia_unregister_driver(&nsp_driver);
}


module_init(nsp_cs_init);
module_exit(nsp_cs_exit);

/* end */
